"""Defines the linter class."""

import fnmatch
import logging
import os
import time
from collections.abc import Iterator, Sequence
from typing import TYPE_CHECKING, Optional, Union, cast

import regex
from tqdm import tqdm

from sqlfluff.core.config import FluffConfig, progress_bar_configuration
from sqlfluff.core.errors import (
    SQLBaseError,
    SQLFluffSkipFile,
    SQLLexError,
    SQLLintError,
    SQLParseError,
    SQLTemplaterError,
)
from sqlfluff.core.formatter import FormatterInterface
from sqlfluff.core.helpers.file import get_encoding
from sqlfluff.core.linter.common import (
    ParsedString,
    ParsedVariant,
    RenderedFile,
    RuleTuple,
)
from sqlfluff.core.linter.discovery import paths_from_path
from sqlfluff.core.linter.fix import apply_fixes, compute_anchor_edit_info
from sqlfluff.core.linter.linted_dir import LintedDir
from sqlfluff.core.linter.linted_file import (
    TMP_PRS_ERROR_TYPES,
    FileTimings,
    LintedFile,
)
from sqlfluff.core.linter.linting_result import LintingResult
from sqlfluff.core.linter.patch import generate_source_patches, merge_source_patches
from sqlfluff.core.parser import Lexer, Parser
from sqlfluff.core.parser.segments.base import BaseSegment, SourceFix
from sqlfluff.core.rules import BaseRule, RulePack, get_ruleset
from sqlfluff.core.rules.fix import LintFix
from sqlfluff.core.rules.noqa import IgnoreMask

if TYPE_CHECKING:  # pragma: no cover
    from sqlfluff.core.dialects import Dialect
    from sqlfluff.core.parser.segments.meta import MetaSegment
    from sqlfluff.core.templaters import RawTemplater, TemplatedFile


RuleTimingsType = list[tuple[str, str, float]]

# Instantiate the linter logger
linter_logger: logging.Logger = logging.getLogger("sqlfluff.linter")


class Linter:
    """The interface class to interact with the linter."""

    # Default to allowing process parallelism
    allow_process_parallelism = True

    def __init__(
        self,
        config: Optional[FluffConfig] = None,
        formatter: Optional[FormatterInterface] = None,
        dialect: Optional[str] = None,
        rules: Optional[list[str]] = None,
        user_rules: Optional[list[type[BaseRule]]] = None,
        exclude_rules: Optional[list[str]] = None,
    ) -> None:
        if config and (dialect or rules or exclude_rules):
            raise ValueError(  # pragma: no cover
                "Linter does not support setting both `config` and any of "
                "`dialect`, `rules` or `exclude_rules`. The latter are "
                "provided as convenience methods to avoid needing to "
                "set the `config` object. If using `config`, please "
                "provide all the other values within that object."
            )
        # Use the provided config or create one from the kwargs.
        self.config = config or FluffConfig.from_kwargs(
            dialect=dialect,
            rules=rules,
            exclude_rules=exclude_rules,
            # Don't require a dialect to be provided yet. Defer this until we
            # are actually linting something, since the directory we are linting
            # from may provide additional configuration, including a dialect.
            require_dialect=False,
        )
        # Get the dialect and templater
        self.dialect: "Dialect" = cast("Dialect", self.config.get("dialect_obj"))
        self.templater: "RawTemplater" = cast(
            "RawTemplater", self.config.get("templater_obj")
        )
        # Store the formatter for output
        self.formatter = formatter
        # Store references to user rule classes
        self.user_rules = user_rules or []

    def get_rulepack(self, config: Optional[FluffConfig] = None) -> RulePack:
        """Get hold of a set of rules."""
        rs = get_ruleset()
        # Register any user rules
        for rule in self.user_rules:
            rs.register(rule)
        cfg = config or self.config
        return rs.get_rulepack(config=cfg)

    def rule_tuples(self) -> list[RuleTuple]:
        """A simple pass through to access the rule tuples of the rule set."""
        rs = self.get_rulepack()
        return [
            RuleTuple(rule.code, rule.name, rule.description, rule.groups, rule.aliases)
            for rule in rs.rules
        ]

    # #### Static methods
    # These are the building blocks of the linting process.

    @staticmethod
    def load_raw_file_and_config(
        fname: str, root_config: FluffConfig
    ) -> tuple[str, FluffConfig, str]:
        """Load a raw file and the associated config."""
        file_config = root_config.make_child_from_path(fname)
        config_encoding: str = file_config.get("encoding", default="autodetect")
        encoding = get_encoding(fname=fname, config_encoding=config_encoding)
        # Check file size before loading.
        limit = file_config.get("large_file_skip_byte_limit")

        if limit:
            # make sure the limit becomes an integer
            try:
                limit = int(limit)
            except ValueError:
                raise ValueError(f"""
                large_file_skip_byte_limit parameter from config
                cannot be converted to integer,
                current value {limit}, type {type(limit)}
                """)
            except TypeError:
                raise TypeError(f"""
                failed to get large_file_skip_byte_limit parameter from config,
                or it is of invalid type {type(limit)}
                """)
            # Get the file size
            file_size = os.path.getsize(fname)
            if file_size > limit:
                raise SQLFluffSkipFile(
                    f"Length of file {fname!r} is {file_size} bytes which is over "
                    f"the limit of {limit} bytes. Skipping to avoid parser lock. "
                    "Users can increase this limit in their config by setting the "
                    "'large_file_skip_byte_limit' value, or disable by setting it "
                    "to zero."
                )
        with open(fname, encoding=encoding, errors="backslashreplace") as target_file:
            raw_file = target_file.read()
        # Scan the raw file for config commands.
        file_config.process_raw_file_for_config(raw_file, fname)
        # Return the raw file and config
        return raw_file, file_config, encoding

    @staticmethod
    def _normalise_newlines(string: str) -> str:
        """Normalise newlines to unix-style line endings."""
        return regex.sub(r"\r\n|\r", "\n", string)

    @staticmethod
    def _lex_templated_file(
        templated_file: "TemplatedFile", config: FluffConfig
    ) -> tuple[Optional[Sequence[BaseSegment]], list[SQLLexError]]:
        """Lex a templated file."""
        violations = []
        linter_logger.info("LEXING RAW (%s)", templated_file.fname)
        # Get the lexer
        lexer = Lexer(config=config)
        # Lex the file and log any problems
        try:
            segments, lex_vs = lexer.lex(templated_file)
            # NOTE: There will always be segments, even if it's
            # just an end of file marker.
            assert segments, "The token sequence should never be empty."
            # We might just get the violations as a list
            violations += lex_vs
            linter_logger.info("Lexed segments: %s", [seg.raw for seg in segments])
        except SQLLexError as err:  # pragma: no cover
            linter_logger.info("LEXING FAILED! (%s): %s", templated_file.fname, err)
            violations.append(err)
            return None, violations

        # Check that we've got sensible indentation from the lexer.
        # We might need to suppress if it's a complicated file.
        templating_blocks_indent = config.get("template_blocks_indent", "indentation")
        if isinstance(templating_blocks_indent, str):
            force_block_indent = templating_blocks_indent.lower().strip() == "force"
        else:
            force_block_indent = False
        templating_blocks_indent = bool(templating_blocks_indent)
        # If we're forcing it through we don't check.
        if templating_blocks_indent and not force_block_indent:
            indent_balance = sum(getattr(elem, "indent_val", 0) for elem in segments)
            if indent_balance != 0:  # pragma: no cover
                linter_logger.debug(
                    "Indent balance test failed for %r. Template indents will not be "
                    "linted for this file.",
                    templated_file.fname,
                )
                # Don't enable the templating blocks.
                templating_blocks_indent = False

        # The file will have been lexed without config, so check all indents
        # are enabled.
        new_segments = []
        for segment in segments:
            if segment.is_meta:
                meta_segment = cast("MetaSegment", segment)
                if meta_segment.indent_val != 0:
                    # Don't allow it if we're not linting templating block indents.
                    if not templating_blocks_indent:
                        continue  # pragma: no cover
            new_segments.append(segment)

        # Return new buffer
        return new_segments, violations

    @staticmethod
    def _parse_tokens(
        tokens: Sequence[BaseSegment],
        config: FluffConfig,
        fname: Optional[str] = None,
        parse_statistics: bool = False,
    ) -> tuple[Optional[BaseSegment], list[SQLParseError]]:
        max_parse_nodes = config.get("max_parse_nodes")
        assert isinstance(max_parse_nodes, int)
        if max_parse_nodes > 0 and len(tokens) > max_parse_nodes:
            anchor = next((seg for seg in tokens if seg.is_code), None)
            err = SQLParseError(
                description=(
                    f"Maximum parse node count exceeded (limit {max_parse_nodes}). "
                    "This may indicate unusually large SQL or a malicious input."
                ),
                segment=anchor,
            )
            linter_logger.info("PARSING SKIPPED! : %s", err)
            return None, [err]

        # Use Rust parser if configured (experimental)
        use_rust = config.get_section(["core", "use_rust_parser"])

        # Determine whether to use Rust parser
        # - "auto": Use if available, fall back silently if not
        # - True/"True"/"true"/1: Force usage, warn if unavailable
        # - False/"False"/"false"/0: Never use
        use_rust_parser = False
        warn_if_unavailable = False

        if use_rust in ("auto", "Auto", "AUTO"):
            # Auto mode: try to use Rust, but don't warn if unavailable
            use_rust_parser = True
            warn_if_unavailable = False
        elif use_rust in (True, "True", "true", 1):
            # Explicit True: use Rust and warn if unavailable
            use_rust_parser = True
            warn_if_unavailable = True

        if use_rust_parser:
            from sqlfluff.core.parser.rust_parser import RustParser

            if RustParser is not None:
                parser: Union[Parser, "RustParser"] = RustParser(config=config)
                linter_logger.info("Using Rust parser (experimental)")
            else:
                if warn_if_unavailable:
                    linter_logger.warning(
                        "use_rust_parser=True but sqlfluffrs not available. "
                        "Falling back to Python parser. Build with: "
                        "cd sqlfluffrs && maturin develop --features python"
                    )
                parser = Parser(config=config)
        else:
            parser = Parser(config=config)
        violations = []
        # Parse the file and log any problems
        try:
            parsed: Optional[BaseSegment] = parser.parse(
                # Regardless of how the sequence was passed in, we should
                # coerce it to a tuple here, before we head deeper into
                # the parsing process.
                tuple(tokens),
                fname=fname,
                parse_statistics=parse_statistics,
            )
        except SQLParseError as err:
            if err.segment is None:
                anchor = next((seg for seg in tokens if seg.is_code), None)
                if anchor is not None:
                    err = SQLParseError(
                        description=err.description,
                        segment=anchor,
                        ignore=err.ignore,
                        fatal=err.fatal,
                        warning=err.warning,
                    )
            linter_logger.info("PARSING FAILED! : %s", err)
            violations.append(err)
            return None, violations

        if parsed is None:  # pragma: no cover
            return None, violations

        if linter_logger.isEnabledFor(logging.INFO):
            linter_logger.info("\n###\n#\n# {}\n#\n###".format("Parsed Tree:"))
            linter_logger.info("\n" + parsed.stringify())
        # We may succeed parsing, but still have unparsable segments. Extract them
        # here.
        for unparsable in parsed.iter_unparsables():
            # No exception has been raised explicitly, but we still create one here
            # so that we can use the common interface
            assert unparsable.pos_marker
            violations.append(
                SQLParseError(
                    "Line {0[0]}, Position {0[1]}: Found unparsable section: "
                    "{1!r}".format(
                        unparsable.pos_marker.working_loc,
                        (
                            unparsable.raw
                            if len(unparsable.raw) < 40
                            else unparsable.raw[:40] + "..."
                        ),
                    ),
                    segment=unparsable,
                )
            )
            if linter_logger.isEnabledFor(logging.INFO):
                linter_logger.info("Found unparsable segment...")
                linter_logger.info(unparsable.stringify())
        return parsed, violations

    @staticmethod
    def remove_templated_errors(
        linting_errors: list[SQLBaseError],
    ) -> list[SQLBaseError]:
        """Filter a list of lint errors, removing those from the templated slices."""
        # Filter out any linting errors in templated sections if relevant.
        result: list[SQLBaseError] = []
        for e in linting_errors:
            if isinstance(e, SQLLintError):
                assert e.segment.pos_marker
                if (
                    # Is it in a literal section?
                    e.segment.pos_marker.is_literal()
                    # Is it in placeholder-tagged template slices that are
                    # semantically literal SQL (e.g. ${catalog})?
                    or Linter._is_semantically_literal_templated(e)
                    # Is it a rule that is designed to work on templated sections?
                    or e.rule.targets_templated
                ):
                    result.append(e)
            else:
                # If it's another type, just keep it. (E.g. SQLParseError from
                # malformed "noqa" comment).
                result.append(e)
        return result

    @staticmethod
    def _is_semantically_literal_templated(error: SQLLintError) -> bool:
        """Whether all non-literal raw slices in the anchor are semantically literal."""
        assert error.segment.pos_marker
        source_slice = error.segment.pos_marker.source_slice
        non_literal_slices = []
        for raw_slice in error.segment.pos_marker.templated_file.raw_sliced:
            if raw_slice.end_source_idx() <= source_slice.start:
                continue
            if raw_slice.source_idx >= source_slice.stop:
                break
            if raw_slice.slice_type != "literal":
                non_literal_slices.append(raw_slice)
        return bool(non_literal_slices) and all(
            raw_slice.tag == "literal" for raw_slice in non_literal_slices
        )

    @staticmethod
    def _report_conflicting_fixes_same_anchor(message: str) -> None:  # pragma: no cover
        # This function exists primarily in order to let us monkeypatch it at
        # runtime (replacing it with a function that raises an exception).
        linter_logger.critical(message)

    @staticmethod
    def _warn_unfixable(code: str) -> None:
        linter_logger.warning(
            f"One fix for {code} not applied, it would re-cause the same error."
        )

    # ### Class Methods
    # These compose the base static methods into useful recipes.

    @classmethod
    def parse_rendered(
        cls,
        rendered: RenderedFile,
        parse_statistics: bool = False,
    ) -> ParsedString:
        """Parse a rendered file."""
        tokens: Optional[Sequence[BaseSegment]]
        parsed_variants: list[ParsedVariant] = []
        _lexing_time = 0.0
        _parsing_time = 0.0

        for idx, variant in enumerate(rendered.templated_variants):
            t0 = time.monotonic()
            linter_logger.info("Parse Rendered. Lexing Variant %s", idx)
            tokens, lex_errors = cls._lex_templated_file(variant, rendered.config)
            t1 = time.monotonic()
            linter_logger.info("Parse Rendered. Parsing Variant %s", idx)
            if tokens:
                parsed, parse_errors = cls._parse_tokens(
                    tokens,
                    rendered.config,
                    fname=rendered.fname,
                    parse_statistics=parse_statistics,
                )
            else:  # pragma: no cover
                parsed = None
                parse_errors = []
            _lt = t1 - t0
            _pt = time.monotonic() - t1
            linter_logger.info(
                "Parse Rendered. Variant %s. Lex in %s. Parse in %s.", idx, _lt, _pt
            )
            parsed_variants.append(
                ParsedVariant(
                    variant,
                    parsed,
                    lex_errors,
                    parse_errors,
                )
            )
            _lexing_time += _lt
            _parsing_time += _pt

        time_dict = {
            **rendered.time_dict,
            "lexing": _lexing_time,
            "parsing": _parsing_time,
        }
        return ParsedString(
            parsed_variants=parsed_variants,
            templating_violations=rendered.templater_violations,
            time_dict=time_dict,
            config=rendered.config,
            fname=rendered.fname,
            source_str=rendered.source_str,
        )

    @classmethod
    def lint_fix_parsed(
        cls,
        tree: BaseSegment,
        config: FluffConfig,
        rule_pack: RulePack,
        fix: bool = False,
        fname: Optional[str] = None,
        templated_file: Optional["TemplatedFile"] = None,
        formatter: Optional[FormatterInterface] = None,
    ) -> tuple[BaseSegment, list[SQLBaseError], Optional[IgnoreMask], RuleTimingsType]:
        """Lint and optionally fix a tree object."""
        # Keep track of the linting errors on the very first linter pass. The
        # list of issues output by "lint" and "fix" only includes issues present
        # in the initial SQL code, EXCLUDING any issues that may be created by
        # the fixes themselves.
        initial_linting_errors = []
        # A placeholder for the fixes we had on the previous loop
        last_fixes: Optional[list[LintFix]] = None
        # Keep a set of previous versions to catch infinite loops.
        previous_versions: set[tuple[str, tuple["SourceFix", ...]]] = {(tree.raw, ())}
        # Keep a buffer for recording rule timings.
        rule_timings: RuleTimingsType = []

        # If we are fixing then we want to loop up to the runaway_limit, otherwise just
        # once for linting.
        loop_limit = config.get("runaway_limit") if fix else 1

        # Dispatch the output for the lint header
        if formatter:
            formatter.dispatch_lint_header(
                fname or "<filename>", sorted(rule_pack.codes())
            )

        # Look for comment segments which might indicate lines to ignore.
        disable_noqa_except: Optional[str] = config.get("disable_noqa_except")
        if not config.get("disable_noqa") or disable_noqa_except:
            allowed_rules_ref_map = cls.allowed_rule_ref_map(
                rule_pack.reference_map, disable_noqa_except
            )
            ignore_mask, ivs = IgnoreMask.from_tree(tree, allowed_rules_ref_map)
            initial_linting_errors += ivs
        else:
            ignore_mask = None

        save_tree = tree
        # There are two phases of rule running.
        # 1. The main loop is for most rules. These rules are assumed to
        # interact and cause a cascade of fixes requiring multiple passes.
        # These are run the `runaway_limit` number of times (default 10).
        # 2. The post loop is for post-processing rules, not expected to trigger
        # any downstream rules, e.g. capitalization fixes. They are run on the
        # first loop and then twice at the end (once to fix, and once again to
        # check result of fixes), but not in the intervening loops.
        phases = ["main"]
        if fix:
            phases.append("post")
        for phase in phases:
            if len(phases) > 1:
                rules_this_phase = [
                    rule for rule in rule_pack.rules if rule.lint_phase == phase
                ]
            else:
                rules_this_phase = rule_pack.rules
            for loop in range(loop_limit if phase == "main" else 2):

                def is_first_linter_pass() -> bool:
                    return phase == phases[0] and loop == 0

                # Additional newlines are to assist in scanning linting loops
                # during debugging.
                linter_logger.info(
                    f"\n\nEntering linter phase {phase}, loop {loop + 1}/{loop_limit}\n"
                )
                changed = False

                if is_first_linter_pass():
                    # In order to compute initial_linting_errors correctly, need
                    # to run all rules on the first loop of the main phase.
                    rules_this_phase = rule_pack.rules
                progress_bar_crawler = tqdm(
                    rules_this_phase,
                    desc="lint by rules",
                    leave=False,
                    disable=progress_bar_configuration.disable_progress_bar,
                )

                for crawler in progress_bar_crawler:
                    # Performance: After first loop pass, skip rules that don't
                    # do fixes. Any results returned won't be seen by the user
                    # anyway (linting errors ADDED by rules changing SQL, are
                    # not reported back to the user - only initial linting errors),
                    # so there's absolutely no reason to run them.
                    if (
                        fix
                        and not is_first_linter_pass()
                        and not crawler.is_fix_compatible
                    ):
                        continue

                    progress_bar_crawler.set_description(f"rule {crawler.code}")
                    t0 = time.monotonic()

                    # fixes should be a dict {} with keys edit, delete, create
                    # delete is just a list of segments to delete
                    # edit and create are list of tuples. The first element is
                    # the "anchor", the segment to look for either to edit or to
                    # insert BEFORE. The second is the element to insert or create.
                    linting_errors, _, fixes, _ = crawler.crawl(
                        tree,
                        dialect=config.get("dialect_obj"),
                        fix=fix,
                        templated_file=templated_file,
                        ignore_mask=ignore_mask,
                        fname=fname,
                        config=config,
                    )
                    if is_first_linter_pass():
                        initial_linting_errors += linting_errors

                    if fix and fixes:
                        linter_logger.info(f"Applying Fixes [{crawler.code}]: {fixes}")
                        # Do some sanity checks on the fixes before applying.
                        anchor_info = compute_anchor_edit_info(fixes)
                        if any(
                            not info.is_valid for info in anchor_info.values()
                        ):  # pragma: no cover
                            message = (
                                f"Rule {crawler.code} returned conflicting "
                                "fixes with the same anchor. This is only "
                                "supported for create_before+create_after, so "
                                "the fixes will not be applied. "
                            )
                            for uuid, info in anchor_info.items():
                                if not info.is_valid:
                                    message += f"\n{uuid}:"
                                    for _fix in info.fixes:
                                        message += f"\n    {_fix}"
                            cls._report_conflicting_fixes_same_anchor(message)
                            for lint_result in linting_errors:
                                lint_result.fixes = []
                        elif fixes == last_fixes:
                            # If we generate the same fixes two times in a row,
                            # that means we're in a loop, and we want to stop.
                            # (Fixes should address issues, hence different
                            # and/or fewer fixes next time.)
                            # This is most likely because fixes could not be safely
                            # applied last time, so we should stop gracefully.
                            linter_logger.debug(
                                f"Fixes generated for {crawler.code} are the same as "
                                "the previous pass. Assuming that we cannot apply them "
                                "safely. Passing gracefully."
                            )
                        else:
                            # This is the happy path. We have fixes, now we want to
                            # apply them.
                            last_fixes = fixes
                            new_tree, _, _, _valid = apply_fixes(
                                tree,
                                config.get("dialect_obj"),
                                crawler.code,
                                anchor_info,
                                fix_even_unparsable=config.get("fix_even_unparsable"),
                                max_parse_depth=config.get("max_parse_depth"),
                                max_parse_nodes=config.get("max_parse_nodes"),
                            )

                            # Check for infinite loops. We use a combination of the
                            # fixed templated file and the list of source fixes to
                            # apply.
                            loop_check_tuple = (
                                new_tree.raw,
                                tuple(new_tree.source_fixes),
                            )
                            # Was anything actually applied? If not, then the fixes we
                            # had cannot be safely applied and we should stop trying.
                            if loop_check_tuple == (tree.raw, tuple(tree.source_fixes)):
                                linter_logger.debug(
                                    f"Fixes for {crawler.code} could not be safely be "
                                    "applied. Likely due to initially unparsable file."
                                )
                            elif not _valid:
                                # The fixes result in an invalid file. Don't apply
                                # the fix and skip onward. Show a warning.
                                linter_logger.warning(
                                    f"Fixes for {crawler.code} not applied, as it "
                                    "would result in an unparsable file. Please "
                                    "report this as a bug with a minimal query "
                                    "which demonstrates this warning."
                                )
                            elif loop_check_tuple not in previous_versions:
                                # We've not seen this version of the file so
                                # far. Continue.
                                tree = new_tree
                                previous_versions.add(loop_check_tuple)
                                changed = True
                                continue
                            else:
                                # Applying these fixes took us back to a state
                                # which we've seen before. We're in a loop, so
                                # we want to stop.
                                cls._warn_unfixable(crawler.code)

                    # Record rule timing
                    rule_timings.append(
                        (crawler.code, crawler.name, time.monotonic() - t0)
                    )

                if fix and not changed:
                    # We did not change the file. Either the file is clean (no
                    # fixes), or any fixes which are present will take us back
                    # to a previous state.
                    linter_logger.info(
                        f"Fix loop complete for {phase} phase. Stability "
                        f"achieved after {loop}/{loop_limit} loops."
                    )
                    break
            else:
                if fix:
                    # The linter loop hit the limit before reaching a stable point
                    # (i.e. free of lint errors). If this happens, it's usually
                    # because one or more rules produced fixes which did not address
                    # the original issue **or** created new issues.
                    linter_logger.warning(
                        f"Loop limit on fixes reached [{loop_limit}]."
                    )

                    # Discard any fixes for the linting errors, since they caused a
                    # loop. IMPORTANT: By doing this, we are telling SQLFluff that
                    # these linting errors are "unfixable". This is important,
                    # because when "sqlfluff fix" encounters unfixable lint errors,
                    # it exits with a "failure" exit code, which is exactly what we
                    # want in this situation. (Reason: Although this is more of an
                    # internal SQLFluff issue, users deserve to know about it,
                    # because it means their file(s) weren't fixed.
                    for violation in initial_linting_errors:
                        if isinstance(violation, SQLLintError):
                            violation.fixes = []

                    # Return the original parse tree, before any fixes were applied.
                    # Reason: When the linter hits the loop limit, the file is often
                    # messy, e.g. some of the fixes were applied repeatedly, possibly
                    # other weird things. We don't want the user to see this junk!
                    return save_tree, initial_linting_errors, ignore_mask, rule_timings

        if config.get("ignore_templated_areas", default=True):
            initial_linting_errors = cls.remove_templated_errors(initial_linting_errors)

        if linter_logger.isEnabledFor(logging.INFO):
            linter_logger.info("\n###\n#\n# {}\n#\n###".format("Fixed Tree:"))
            linter_logger.info("\n" + tree.stringify())

        return tree, initial_linting_errors, ignore_mask, rule_timings

    @classmethod
    def lint_parsed(
        cls,
        parsed: ParsedString,
        rule_pack: RulePack,
        fix: bool = False,
        formatter: Optional[FormatterInterface] = None,
        encoding: str = "utf8",
    ) -> LintedFile:
        """Lint a ParsedString and return a LintedFile."""
        time_dict = parsed.time_dict
        tree: Optional[BaseSegment] = None
        templated_file: Optional[TemplatedFile] = None
        merged_source_patches = None
        t0 = time.monotonic()

        root_variant = parsed.root_variant()
        if not root_variant:
            linter_logger.info(
                "lint_parsed found no valid root variant for %s", parsed.fname
            )

        # Always surface templating violations. If we found a valid root variant,
        # only seed parse/lex violations from that root variant; alternate variants
        # are used to discover extra linting errors, but parse failures in them
        # should not make an otherwise valid file fail linting/fixing.
        violations: list[SQLBaseError] = list(parsed.templating_violations)
        if root_variant:
            violations += root_variant.violations()
        else:
            violations += [
                violation
                for variant in parsed.parsed_variants
                for violation in variant.violations()
            ]

        # If there is a root variant, handle that first.
        if root_variant:
            linter_logger.info("lint_parsed - linting root variant (%s)", parsed.fname)
            assert root_variant.tree  # We just checked this.
            variant_source_patches = []
            (
                fixed_tree,
                initial_linting_errors,
                ignore_mask,
                rule_timings,
            ) = cls.lint_fix_parsed(
                root_variant.tree,
                config=parsed.config,
                rule_pack=rule_pack,
                fix=fix,
                fname=parsed.fname,
                templated_file=root_variant.templated_file,
                formatter=formatter,
            )

            # Set legacy variables for the return payload.
            templated_file = root_variant.templated_file
            tree = fixed_tree
            if fix:
                variant_source_patches.append(
                    generate_source_patches(fixed_tree, root_variant.templated_file)
                )

            # We're only going to return the *initial* errors, rather
            # than any generated during the fixing cycle.
            violations += initial_linting_errors

            # Lint alternate variants (if they exist) so branch-specific violations from
            # templated code are surfaced in the final deduplicated result set.
            for idx, alternate_variant in enumerate(parsed.parsed_variants):
                if alternate_variant is root_variant or not alternate_variant.tree:
                    continue
                linter_logger.info("lint_parsed - linting alt variant (%s)", idx)
                (
                    alt_fixed_tree,
                    alt_linting_errors,
                    _,  # Ignore Mask
                    _,  # Timings
                ) = cls.lint_fix_parsed(
                    alternate_variant.tree,
                    config=parsed.config,
                    rule_pack=rule_pack,
                    fix=fix,
                    fname=parsed.fname,
                    templated_file=alternate_variant.templated_file,
                    formatter=formatter,
                )
                violations += alt_linting_errors
                if fix:
                    variant_source_patches.append(
                        generate_source_patches(
                            alt_fixed_tree,
                            alternate_variant.templated_file,
                        )
                    )

            if fix:
                merged_source_patches = merge_source_patches(variant_source_patches)

        # If no root variant, we should still apply ignores to any parsing
        # or templating fails.
        else:
            rule_timings = []
            disable_noqa_except: Optional[str] = parsed.config.get(
                "disable_noqa_except"
            )
            if parsed.config.get("disable_noqa") and not disable_noqa_except:
                # NOTE: This path is only accessible if there is no valid `tree`
                # which implies that there was a fatal templating fail. Even an
                # unparsable file will still have a valid tree.
                ignore_mask = None
            else:
                # Templating and/or parsing have failed. Look for "noqa"
                # comments (the normal path for identifying these comments
                # requires access to the parse tree, and because of the failure,
                # we don't have a parse tree).
                allowed_rules_ref_map = cls.allowed_rule_ref_map(
                    rule_pack.reference_map, disable_noqa_except
                )
                ignore_mask, ignore_violations = IgnoreMask.from_source_with_dialect(
                    parsed.source_str,
                    parsed.config.get("dialect_obj"),
                    allowed_rules_ref_map,
                )
                violations += ignore_violations

        # Update the timing dict
        time_dict["linting"] = time.monotonic() - t0

        # We process the ignore config here if appropriate
        for violation in violations:
            violation.ignore_if_in(parsed.config.get("ignore"))
            violation.warning_if_in(parsed.config.get("warnings"))

        linted_file = LintedFile(
            parsed.fname,
            # Deduplicate violations
            LintedFile.deduplicate_in_source_space(violations),
            FileTimings(time_dict, rule_timings),
            tree,
            ignore_mask=ignore_mask,
            templated_file=templated_file,
            encoding=encoding,
            source_patches=merged_source_patches,
        )

        # This is the main command line output from linting.
        if formatter:
            formatter.dispatch_file_violations(
                parsed.fname,
                linted_file,
                only_fixable=fix,
                warn_unused_ignores=parsed.config.get("warn_unused_ignores"),
            )

        # Safety flag for unset dialects
        if linted_file.get_violations(
            fixable=True if fix else None, types=SQLParseError
        ):
            if formatter:  # pragma: no cover TODO?
                formatter.dispatch_dialect_warning(
                    # The dialect property is the string, not the dialect object
                    cast(str, parsed.config.get("dialect"))
                )

        return linted_file

    @classmethod
    def allowed_rule_ref_map(
        cls, reference_map: dict[str, set[str]], disable_noqa_except: Optional[str]
    ) -> dict[str, set[str]]:
        """Generate a noqa rule reference map."""
        # disable_noqa_except is not set, return the entire map.
        if not disable_noqa_except:
            return reference_map
        output_map = reference_map
        # Add the special rules so they can be excluded for `disable_noqa_except` usage
        for special_rule in ["PRS", "LXR", "TMP"]:
            output_map[special_rule] = {special_rule}
        # Expand glob usage of rules
        unexpanded_rules = tuple(r.strip() for r in disable_noqa_except.split(","))
        noqa_set = set()
        for r in unexpanded_rules:
            for x in fnmatch.filter(output_map.keys(), r):
                noqa_set |= output_map.get(x, set())
        # Return a new map with only the excluded rules
        return {k: v.intersection(noqa_set) for k, v in output_map.items()}

    @classmethod
    def lint_rendered(
        cls,
        rendered: RenderedFile,
        rule_pack: RulePack,
        fix: bool = False,
        formatter: Optional[FormatterInterface] = None,
    ) -> LintedFile:
        """Take a RenderedFile and return a LintedFile."""
        parsed = cls.parse_rendered(rendered)
        return cls.lint_parsed(
            parsed,
            rule_pack=rule_pack,
            fix=fix,
            formatter=formatter,
            encoding=rendered.encoding,
        )

    # ### Instance Methods
    # These are tied to a specific instance and so are not necessarily
    # safe to use in parallel operations.

    def render_string(
        self, in_str: str, fname: str, config: FluffConfig, encoding: str
    ) -> RenderedFile:
        """Template the file."""
        linter_logger.info("Rendering String [%s] (%s)", self.templater.name, fname)

        # Start the templating timer
        t0 = time.monotonic()

        # Newlines are normalised to unix-style line endings (\n).
        # The motivation is that Jinja normalises newlines during templating and
        # we want consistent mapping between the raw and templated slices.
        in_str = self._normalise_newlines(in_str)

        # Since Linter.__init__() does not require a dialect to be specified,
        # check for one now. (We're processing a string, not a file, so we're
        # not going to pick up a .sqlfluff or other config file to provide a
        # missing dialect at this point.)
        config.verify_dialect_specified()
        if not config.get("templater_obj") == self.templater:
            linter_logger.warning(
                f"Attempt to set templater to {config.get('templater_obj').name} "
                f"failed. Using {self.templater.name} templater. Templater cannot "
                "be set in a .sqlfluff file in a subdirectory of the current "
                "working directory. It can be set in a .sqlfluff in the current "
                "working directory. See Nesting section of the docs for more "
                "details."
            )

        variant_limit = config.get("render_variant_limit")
        templated_variants: list[TemplatedFile] = []
        templater_violations: list[SQLTemplaterError] = []

        try:
            for variant, templater_errs in self.templater.process_with_variants(
                in_str=in_str, fname=fname, config=config, formatter=self.formatter
            ):
                if variant:
                    templated_variants.append(variant)
                # Duplicate templater errors can arise across variants. Final
                # linted output deduplicates these in source space, but the
                # intermediate ParsedString still preserves the per-variant list.
                templater_violations += templater_errs
                if len(templated_variants) >= variant_limit:
                    # Stop if we hit the limit.
                    break
        except SQLTemplaterError as templater_err:
            # Fatal templating error. Capture it and don't generate a variant.
            templater_violations.append(templater_err)
        except SQLFluffSkipFile as skip_file_err:  # pragma: no cover
            linter_logger.warning(str(skip_file_err))

        if not templated_variants:
            linter_logger.info("TEMPLATING FAILED: %s", templater_violations)

        linter_logger.info("Rendered %s variants", len(templated_variants))

        # Record time
        time_dict = {"templating": time.monotonic() - t0}

        return RenderedFile(
            templated_variants,
            templater_violations,
            config,
            time_dict,
            fname,
            encoding,
            in_str,
        )

    def render_file(self, fname: str, root_config: FluffConfig) -> RenderedFile:
        """Load and render a file with relevant config."""
        # Load the raw file.
        raw_file, config, encoding = self.load_raw_file_and_config(fname, root_config)
        # Render the file
        return self.render_string(raw_file, fname, config, encoding)

    def parse_string(
        self,
        in_str: str,
        fname: str = "<string>",
        config: Optional[FluffConfig] = None,
        encoding: str = "utf-8",
        parse_statistics: bool = False,
    ) -> ParsedString:
        """Parse a string."""
        violations: list[SQLBaseError] = []

        # Dispatch the output for the template header (including the config diff)
        if self.formatter:
            self.formatter.dispatch_template_header(fname, self.config, config)

        # Just use the local config from here:
        config = (config or self.config).copy()

        # Scan the raw file for config commands.
        config.process_raw_file_for_config(in_str, fname)
        rendered = self.render_string(in_str, fname, config, encoding)
        violations += rendered.templater_violations

        # Dispatch the output for the parse header
        if self.formatter:
            self.formatter.dispatch_parse_header(fname)

        return self.parse_rendered(rendered, parse_statistics=parse_statistics)

    def fix(
        self,
        tree: BaseSegment,
        config: Optional[FluffConfig] = None,
        fname: Optional[str] = None,
        templated_file: Optional["TemplatedFile"] = None,
    ) -> tuple[BaseSegment, list[SQLBaseError]]:
        """Return the fixed tree and violations from lintfix when we're fixing."""
        config = config or self.config
        rule_pack = self.get_rulepack(config=config)
        fixed_tree, violations, _, _ = self.lint_fix_parsed(
            tree,
            config,
            rule_pack,
            fix=True,
            fname=fname,
            templated_file=templated_file,
            formatter=self.formatter,
        )
        return fixed_tree, violations

    def lint(
        self,
        tree: BaseSegment,
        config: Optional[FluffConfig] = None,
        fname: Optional[str] = None,
        templated_file: Optional["TemplatedFile"] = None,
    ) -> list[SQLBaseError]:
        """Return just the violations from lintfix when we're only linting."""
        config = config or self.config
        rule_pack = self.get_rulepack(config=config)
        _, violations, _, _ = self.lint_fix_parsed(
            tree,
            config,
            rule_pack,
            fix=False,
            fname=fname,
            templated_file=templated_file,
            formatter=self.formatter,
        )
        return violations

    def lint_string(
        self,
        in_str: str = "",
        fname: str = "<string input>",
        fix: bool = False,
        config: Optional[FluffConfig] = None,
        encoding: str = "utf8",
    ) -> LintedFile:
        """Lint a string.

        Returns:
            :obj:`LintedFile`: an object representing that linted file.

        """
        # Sort out config, defaulting to the built in config if no override
        config = config or self.config
        # Parse the string.
        parsed = self.parse_string(
            in_str=in_str,
            fname=fname,
            config=config,
        )
        # Get rules as appropriate
        rule_pack = self.get_rulepack(config=config)
        # Lint the file and return the LintedFile
        return self.lint_parsed(
            parsed,
            rule_pack,
            fix=fix,
            formatter=self.formatter,
            encoding=encoding,
        )

    def lint_string_wrapped(
        self,
        string: str,
        fname: str = "<string input>",
        fix: bool = False,
        stdin_filename: Optional[str] = None,
    ) -> LintingResult:
        """Lint strings directly."""
        result = LintingResult()
        linted_path = LintedDir(fname)
        if stdin_filename:
            # This opens up additional files like `Untitled` in vscode where linting
            # may be desired.
            sql_exts = ("",)
            for _ in paths_from_path(
                stdin_filename, target_file_exts=sql_exts, check_non_existent_file=True
            ):
                linted_path.add(self.lint_string(string, fname=fname, fix=fix))
                # Only return the first item, as stdin should only be one
                break
        else:
            linted_path.add(self.lint_string(string, fname=fname, fix=fix))
        result.add(linted_path)
        result.stop_timer()
        return result

    def lint_path(
        self,
        path: str,
        fix: bool = False,
        ignore_non_existent_files: bool = False,
        ignore_files: bool = True,
        processes: Optional[int] = None,
    ) -> LintedDir:
        """Lint a path."""
        return self.lint_paths(
            (path,), fix, ignore_non_existent_files, ignore_files, processes
        ).paths[0]

    def lint_paths(
        self,
        paths: tuple[str, ...],
        fix: bool = False,
        ignore_non_existent_files: bool = False,
        ignore_files: bool = True,
        processes: Optional[int] = None,
        apply_fixes: bool = False,
        fixed_file_suffix: str = "",
        fix_even_unparsable: bool = False,
        retain_files: bool = True,
    ) -> LintingResult:
        """Lint an iterable of paths."""
        # If no paths specified - assume local
        if not paths:  # pragma: no cover
            paths = (os.getcwd(),)
        # Set up the result to hold what we get back
        result = LintingResult()

        expanded_paths: list[str] = []
        expanded_path_to_linted_dir = {}
        sql_exts = self.config.get("sql_file_exts", default=".sql").lower().split(",")

        for path in paths:
            linted_dir = LintedDir(path, retain_files=retain_files)
            result.add(linted_dir)
            for fname in paths_from_path(
                path,
                ignore_non_existent_files=ignore_non_existent_files,
                ignore_files=ignore_files,
                target_file_exts=sql_exts,
            ):
                expanded_paths.append(fname)
                expanded_path_to_linted_dir[fname] = linted_dir

        files_count = len(expanded_paths)
        if processes is None:
            processes = self.config.get("processes", default=1)
        assert processes is not None
        # Hard set processes to 1 if only 1 file is queued.
        # The overhead will never be worth it with one file.
        if files_count == 1:
            processes = 1

        # to avoid circular import
        from sqlfluff.core.linter.runner import get_runner

        runner, effective_processes = get_runner(
            self,
            self.config,
            processes=processes,
            allow_process_parallelism=self.allow_process_parallelism,
        )

        if self.formatter and effective_processes != 1:
            self.formatter.dispatch_processing_header(effective_processes)

        # Show files progress bar only when there is more than one.
        first_path = expanded_paths[0] if expanded_paths else ""
        progress_bar_files = tqdm(
            total=files_count,
            desc=f"file {first_path}",
            leave=False,
            disable=files_count <= 1 or progress_bar_configuration.disable_progress_bar,
        )

        runner_iterator = runner.run(expanded_paths, fix)
        try:
            for i, linted_file in enumerate(runner_iterator, start=1):
                linted_dir = expanded_path_to_linted_dir[linted_file.path]
                linted_dir.add(linted_file)
                # If any fatal errors, then stop iteration.
                if any(v.fatal for v in linted_file.violations):  # pragma: no cover
                    linter_logger.error("Fatal linting error. Halting further linting.")
                    break

                # If we're applying fixes, then do that here.
                if apply_fixes:
                    num_tmp_prs_errors = linted_file.num_violations(
                        types=TMP_PRS_ERROR_TYPES,
                        filter_ignore=False,
                        filter_warning=False,
                    )
                    if fix_even_unparsable or num_tmp_prs_errors == 0:
                        linted_file.persist_tree(
                            suffix=fixed_file_suffix, formatter=self.formatter
                        )

                # Progress bar for files is rendered only when there is more
                # than one file. Additionally, as it's updated after each loop,
                # we need to get file name from the next loop. This is why
                # `enumerate` starts with `1` and there is `i < len` to not
                # exceed files list length.
                progress_bar_files.update(n=1)
                if i < len(expanded_paths):
                    progress_bar_files.set_description(f"file {expanded_paths[i]}")
        finally:
            progress_bar_files.close()
            runner_close = getattr(runner_iterator, "close", None)
            if runner_close:
                runner_close()

        # Transfer skipped file count from the runner to the result.
        result.files_skipped = runner.skipped_file_count
        result.stop_timer()
        return result

    def parse_path(
        self,
        path: str,
        parse_statistics: bool = False,
    ) -> Iterator[ParsedString]:
        """Parse a path of sql files.

        NB: This a generator which will yield the result of each file
        within the path iteratively.
        """
        sql_exts = self.config.get("sql_file_exts", default=".sql").lower().split(",")
        for fname in paths_from_path(
            path,
            target_file_exts=sql_exts,
        ):
            if self.formatter:
                self.formatter.dispatch_path(path)
            # Load the file with the config and yield the result.
            try:
                raw_file, config, encoding = self.load_raw_file_and_config(
                    fname, self.config
                )
            except SQLFluffSkipFile as s:
                linter_logger.warning(str(s))
                continue
            yield self.parse_string(
                raw_file,
                fname=fname,
                config=config,
                encoding=encoding,
                parse_statistics=parse_statistics,
            )



"""Implementation of Rule ST09."""

from typing import Optional

from sqlfluff.core.dialects.common import (
    AliasInfo,
    get_from_expression_element_alias,
    get_join_clause_aliases,
)
from sqlfluff.core.parser import BaseSegment, SymbolSegment
from sqlfluff.core.rules import BaseRule, LintFix, LintResult, RuleContext
from sqlfluff.core.rules.crawlers import SegmentSeekerCrawler
from sqlfluff.utils.functional import FunctionalContext, Segments


class Rule_ST09(BaseRule):
    """Joins should list the table referenced earlier/later first.

    This rule will break conditions from join clauses down into subconditions
    using the "and" and "or" binary operators.

    Subconditions that are made up of a qualified column reference,
    a comparison operator and another qualified column reference
    are then evaluated to check whether they list the table that was referenced
    earlier - or later, depending on the ``preferred_first_table_in_join_clause``
    configuration.

    Subconditions that do not follow that pattern are ignored by this rule.

    .. note::
       Joins in ``WHERE`` clauses are currently not supported by this rule.

    **Anti-pattern**

    In this example, the tables that were referenced later are listed first
    and the ``preferred_first_table_in_join_clause`` configuration
    is set to ``earlier``.

    .. code-block:: sql

        select
            foo.a,
            foo.b,
            bar.c
        from foo
        left join bar
            -- This subcondition does not list
            -- the table referenced earlier first:
            on bar.a = foo.a
            -- Neither does this subcondition:
            and bar.b = foo.b

    **Best practice**

    List the tables that were referenced earlier first.

    .. code-block:: sql

        select
            foo.a,
            foo.b,
            bar.c
        from foo
        left join bar
            on foo.a = bar.a
            and foo.b = bar.b
    """

    name = "structure.join_condition_order"
    aliases = ()
    groups: tuple[str, ...] = ("all", "structure")
    config_keywords = ["preferred_first_table_in_join_clause"]
    crawl_behaviour = SegmentSeekerCrawler({"from_expression"})
    is_fix_compatible = True

    # Operators that are either commutative (can swap sides without change) or
    # have known inverses (swap sides and flip operator). Directional operators
    # such as ~ (PostgreSQL regex match) are not included and will be ignored.
    _REORDERABLE_OPERATORS: frozenset[str] = frozenset(
        {"=", "!=", "<>", "<=>", "<", ">", "<=", ">="}
    )

    def _eval(self, context: RuleContext) -> Optional[LintResult]:
        """Find rule violations and provide fixes.

        0. Grab all table aliases into a table_aliases list.
        1. Grab all conditions from the different join_on_condition segments.
        2. Break conditions down into subconditions using the "and" and "or"
        binary operators.
        3. Keep subconditions that are made up of a qualified column_reference,
        a comparison_operator and another qualified column_reference segments.
        4. Check whether the table associated with the first column_reference segment
        has a greater index in table_aliases than the second column_reference segment.
        If so, populate the fixes list (lower index instead of greater index
        if preferred_first_table_in_join_clause == "later").
        5.a. If fixes is empty the rule passes.
        5.b. If fixes isn't empty we return a LintResult object with fixable violations.
        """
        self.preferred_first_table_in_join_clause: str

        assert context.segment.is_type("from_expression")

        # STEP 0.
        table_aliases: list[str] = []

        children = FunctionalContext(context).segment.children()

        # we use recursive_crawl to deal with brackets
        join_clauses = children.recursive_crawl("join_clause")

        join_on_conditions = join_clauses.children().recursive_crawl(
            "join_on_condition"
        )

        # we only care about join_on_condition segments
        if len(join_on_conditions) == 0:
            return None

        # the first alias comes from the from clause
        from_expression_alias_info = next(
            get_from_expression_element_alias(
                children.recursive_crawl("from_expression_element")[0],
                context.dialect.name,
            )
        )
        from_expression_alias: str = (
            from_expression_alias_info.segment.raw_normalized(False)
            if from_expression_alias_info.segment
            else from_expression_alias_info.ref_str
        )

        table_aliases.append(from_expression_alias)

        # the rest of the aliases come from the different join clauses
        join_clause_alias_infos: list[AliasInfo] = [
            get_join_clause_aliases(join_clause, context.dialect.name)[0][1]
            for join_clause in [clause for clause in join_clauses]
        ]

        join_clause_aliases = [
            (
                alias_info.segment.raw_normalized(False)
                if alias_info.segment
                else alias_info.ref_str
            )
            for alias_info in join_clause_alias_infos
        ]

        table_aliases += join_clause_aliases

        table_aliases = [alias.upper() for alias in table_aliases]

        # STEP 1.
        conditions: list[list[BaseSegment]] = []

        join_on_condition__expressions = join_on_conditions.children().recursive_crawl(
            "expression"
        )

        for expression in join_on_condition__expressions:
            expression_group = []
            for element in Segments(expression).children():
                if element.type not in ("whitespace", "newline"):
                    expression_group.append(element)
            conditions.append(expression_group)

        # STEP 2.
        subconditions: list[list[list[BaseSegment]]] = []

        for expression_group in conditions:
            subconditions.append(
                self._split_list_by_segment_type(
                    segment_list=expression_group,
                    delimiter_type="binary_operator",
                    delimiters=["and", "or"],
                )
            )

        subconditions_flattened: list[list[BaseSegment]] = [
            item for sublist in subconditions for item in sublist
        ]

        # STEP 3.
        column_operator_column_subconditions: list[list[BaseSegment]] = [
            subcondition
            for subcondition in subconditions_flattened
            if self._is_qualified_column_operator_qualified_column_sequence(
                subcondition
            )
        ]

        # STEP 4.
        fixes: list[LintFix] = []
        anchor_segment = context.segment  # Default anchor

        for subcondition in column_operator_column_subconditions:
            comparison_operator = subcondition[1]
            first_column_reference = subcondition[0]
            second_column_reference = subcondition[2]
            raw_comparison_operators = comparison_operator.get_children(
                "raw_comparison_operator"
            )

            # Skip operators that cannot be safely reordered. Build the
            # operator string from raw_comparison_operator children if
            # present, otherwise fall back to the comparison_operator raw
            # value (e.g. for ~ or <=> which have no raw_comparison_operator
            # children).
            operator_str = (
                "".join(r.raw for r in raw_comparison_operators)
                if raw_comparison_operators
                else comparison_operator.raw.strip()
            )
            if operator_str not in self._REORDERABLE_OPERATORS:
                continue

            first_table_seg = first_column_reference.get_child(
                "naked_identifier", "quoted_identifier"
            )
            second_table_seg = second_column_reference.get_child(
                "naked_identifier", "quoted_identifier"
            )
            assert first_table_seg and second_table_seg
            first_table = first_table_seg.raw_normalized(False).upper()
            second_table = second_table_seg.raw_normalized(False).upper()

            # if we swap the two column references around the comparison operator
            # we might have to replace the comparison operator with a different one
            raw_comparison_operator_opposites = {"<": ">", ">": "<"}

            # there seem to be edge cases where either the first table or the second
            # table is not in table_aliases, in which case we cannot provide any fix
            if first_table not in table_aliases or second_table not in table_aliases:
                continue

            if (
                table_aliases.index(first_table) > table_aliases.index(second_table)
                and self.preferred_first_table_in_join_clause == "earlier"
            ) or (
                table_aliases.index(first_table) < table_aliases.index(second_table)
                and self.preferred_first_table_in_join_clause == "later"
            ):
                # Use the first column reference as anchor if it has a literal
                # position marker. This ensures the violation is anchored to
                # a literal segment which won't be filtered out in templated
                # code.
                if (
                    not fixes
                    and first_column_reference.pos_marker
                    and first_column_reference.pos_marker.is_literal()
                ):
                    anchor_segment = first_column_reference

                fixes = (
                    fixes
                    + [
                        LintFix.replace(
                            first_column_reference,
                            [second_column_reference],
                        )
                    ]
                    + [
                        LintFix.replace(
                            second_column_reference,
                            [first_column_reference],
                        )
                    ]
                    + (
                        [
                            LintFix.replace(
                                raw_comparison_operators[0],
                                [
                                    SymbolSegment(
                                        raw=raw_comparison_operator_opposites[
                                            raw_comparison_operators[0].raw
                                        ],
                                        type="raw_comparison_operator",
                                    )
                                ],
                            )
                        ]
                        if raw_comparison_operators
                        and raw_comparison_operators[0].raw
                        in raw_comparison_operator_opposites
                        and [r.raw for r in raw_comparison_operators] != ["<", ">"]
                        else []
                    )
                )

        # STEP 5.a.
        if not fixes:
            return None

        # STEP 5.b.
        else:
            return LintResult(
                anchor=anchor_segment,
                fixes=fixes,
                description=(
                    "Joins should list the table referenced "
                    f"{self.preferred_first_table_in_join_clause} first."
                ),
            )

    @staticmethod
    def _split_list_by_segment_type(
        segment_list: list[BaseSegment], delimiter_type: str, delimiters: list[str]
    ) -> list:
        # Break down a list into multiple sub-lists using a set of delimiters
        delimiters = [delimiter.upper() for delimiter in delimiters]
        new_list = []
        sub_list = []
        for i in range(len(segment_list)):
            if i == len(segment_list) - 1:
                sub_list.append(segment_list[i])
                new_list.append(sub_list)
            elif (
                segment_list[i].type == delimiter_type
                and segment_list[i].raw_upper in delimiters
            ):
                new_list.append(sub_list)
                sub_list = []
            else:
                sub_list.append(segment_list[i])

        return new_list

    @staticmethod
    def _is_qualified_column_operator_qualified_column_sequence(
        segment_list: list[BaseSegment],
    ) -> bool:
        # Check if list is made up of a qualified column_reference segment,
        # a comparison_operator segment and another qualified column_reference segment
        if len(segment_list) != 3:
            return False
        if (
            segment_list[0].type == "column_reference"
            and "dot" in segment_list[0].direct_descendant_type_set
            and segment_list[1].type == "comparison_operator"
            and segment_list[2].type == "column_reference"
            and "dot" in segment_list[2].direct_descendant_type_set
        ):
            return True
        return False



"""Implementation of Rule CP03."""

from typing import Optional

from sqlfluff.core.parser.segments.base import BaseSegment
from sqlfluff.core.rules import LintFix, LintResult, RuleContext
from sqlfluff.core.rules.crawlers import SegmentSeekerCrawler
from sqlfluff.rules.capitalisation.CP01 import Rule_CP01

try:
    import sqlfluffrs

    _HAS_SQLFLUFFRS = True
except ImportError:  # pragma: no cover
    _HAS_SQLFLUFFRS = False


class Rule_CP03(Rule_CP01):
    """Inconsistent capitalisation of function names.

    **Anti-pattern**

    In this example, the two ``SUM`` functions don't have the same capitalisation.

    .. code-block:: sql

        SELECT
            sum(a) AS aa,
            SUM(b) AS bb
        FROM foo

    **Best practice**

    Make the case consistent.

    .. code-block:: sql

        SELECT
            sum(a) AS aa,
            sum(b) AS bb
        FROM foo

    """

    name = "capitalisation.functions"
    aliases = ("L030",)
    is_fix_compatible = True

    crawl_behaviour = SegmentSeekerCrawler(
        {"function_name_identifier", "bare_function"}
    )
    _exclude_types = ()
    _exclude_parent_types = ()

    config_keywords = [
        "extended_capitalisation_policy",
        "ignore_words",
        "ignore_words_regex",
    ]
    _description_elem = "Function names"

    def _eval_rust(self, context: RuleContext) -> Optional[list[LintResult]]:
        """Rust-native CP03 detection over the arena.

        Runs the whole detection in Rust (`sqlfluffrs.cp03_violations` over
        `root._rs_tree`) and maps each `(leaf_index, fixed_raw)` back to its
        Python segment to emit a standard `LintFix`. Returns ``None`` (Python
        fallback) when the Rust extension/arena is unavailable, or for a regex
        word-ignore. See ``BaseRule._eval_rust``/``Rule_CP01._eval_rust``.

        Deliberately its own override (not shared dispatch through CP01's) so
        there's nothing for a future capitalisation subclass to hijack, same as
        CP04's.
        """
        if not _HAS_SQLFLUFFRS:
            return None  # pragma: no cover
        if self.name != "capitalisation.functions":
            # Defensive only: no rule currently subclasses CP03, so this is
            # unreachable today — guards against a future subclass silently
            # inheriting this override the way pre-fix CP04 inherited CP01's.
            return None  # pragma: no cover
        if getattr(self, "ignore_words_regex", None):
            return None
        policy = str(getattr(self, "extended_capitalisation_policy"))
        return self._eval_rust_capitalisation(
            context, sqlfluffrs.cp03_violations, policy
        )

    def _get_fix(self, segment: BaseSegment, fixed_raw: str) -> LintFix:
        return super()._get_fix(segment, fixed_raw)



"""The StarRocks dialect.

This dialect extends MySQL grammar with specific StarRocks syntax features.
"""

from sqlfluff.core.dialects import load_raw_dialect
from sqlfluff.core.parser import (
    BaseSegment,
    Bracketed,
    CodeSegment,
    Dedent,
    Delimited,
    ImplicitIndent,
    Matchable,
    MultiStringParser,
    OneOf,
    OptionallyBracketed,
    Ref,
    SegmentGenerator,
    Sequence,
    StringParser,
    SymbolSegment,
)
from sqlfluff.dialects import dialect_ansi as ansi
from sqlfluff.dialects import dialect_mysql as mysql
from sqlfluff.dialects.dialect_starrocks_keywords import (
    starrocks_reserved_keywords,
    starrocks_unreserved_keywords,
)

mysql_dialect = load_raw_dialect("mysql")
starrocks_dialect = mysql_dialect.copy_as(
    "starrocks",
    formatted_name="StarRocks",
    docstring="""**Default Casing**: ``lowercase``

    **Quotes**: String Literals: ``''``, ``"``,
    Identifiers: |back_quotes|.

    The dialect for `StarRocks <https://www.starrocks.io/>`_.""",
)
starrocks_dialect.update_keywords_set_from_multiline_string(
    "unreserved_keywords", starrocks_unreserved_keywords
)
starrocks_dialect.sets("reserved_keywords").clear()
starrocks_dialect.update_keywords_set_from_multiline_string(
    "reserved_keywords", starrocks_reserved_keywords
)


# Add angle bracket pairs for complex type support (ARRAY, MAP, STRUCT)
starrocks_dialect.bracket_sets("angle_bracket_pairs").update(
    [
        ("angle", "StartAngleBracketSegment", "EndAngleBracketSegment", False),
    ]
)

# Add the engine types set
starrocks_dialect.sets("engine_types").update(
    ["olap", "mysql", "elasticsearch", "hive", "hudi", "iceberg", "jdbc"]
)


starrocks_dialect.add(
    StartAngleBracketSegment=StringParser(
        "<", SymbolSegment, type="start_angle_bracket"
    ),
    EndAngleBracketSegment=StringParser(">", SymbolSegment, type="end_angle_bracket"),
    EngineTypeSegment=SegmentGenerator(
        lambda dialect: MultiStringParser(
            dialect.sets("engine_types"),
            CodeSegment,
            type="engine_type",
        )
    ),
)


class ArrayTypeSegment(ansi.ArrayTypeSegment):
    """ARRAY<type> data type for StarRocks.

    https://docs.starrocks.io/docs/sql-reference/data-types/semi_structured/Array/
    """

    type = "array_type"
    match_grammar = Sequence(
        "ARRAY",
        Bracketed(
            Ref("DatatypeSegment"),
            bracket_type="angle",
            bracket_pairs_set="angle_bracket_pairs",
        ),
    )


class DatatypeSegment(mysql.DatatypeSegment):
    """Data type segment extended with StarRocks complex types (ARRAY, MAP, STRUCT).

    https://docs.starrocks.io/docs/sql-reference/data-types/semi_structured/Array/
    https://docs.starrocks.io/docs/sql-reference/data-types/semi_structured/Map/
    https://docs.starrocks.io/docs/sql-reference/data-types/semi_structured/Struct/
    """

    type = "data_type"
    match_grammar = mysql.DatatypeSegment.match_grammar.copy(
        insert=[
            Ref("ArrayTypeSegment"),
            Sequence(
                "MAP",
                Bracketed(
                    Sequence(
                        Ref("DatatypeSegment"),
                        Ref("CommaSegment"),
                        Ref("DatatypeSegment"),
                    ),
                    bracket_pairs_set="angle_bracket_pairs",
                    bracket_type="angle",
                ),
            ),
            Sequence(
                "STRUCT",
                Bracketed(
                    Delimited(
                        Sequence(
                            Ref("SingleIdentifierGrammar"),
                            Ref("DatatypeSegment"),
                        ),
                    ),
                    bracket_pairs_set="angle_bracket_pairs",
                    bracket_type="angle",
                ),
            ),
        ]
    )


class CreateTableStatementSegment(mysql.CreateTableStatementSegment):
    """A `CREATE TABLE` statement.

    StarRocks-specific version that handles:
    - Different ENGINE types
    - Storage and bucketing options
    - Specific partition syntax
    - Specific table properties
    """

    type = "create_table_statement"
    match_grammar = Sequence(
        "CREATE",
        Ref("OrReplaceGrammar", optional=True),
        Ref.keyword("EXTERNAL", optional=True),
        Ref.keyword("TEMPORARY", optional=True),
        "TABLE",
        Ref("IfNotExistsGrammar", optional=True),
        Ref("TableReferenceSegment"),
        OneOf(
            # Standard column definitions
            Sequence(
                Bracketed(
                    Delimited(
                        OneOf(
                            Ref("TableConstraintSegment"),
                            Ref("ColumnDefinitionSegment"),
                            Ref("IndexDefinitionSegment"),
                        ),
                    )
                ),
                # StarRocks specific
                Sequence(
                    "ENGINE",
                    Ref("EqualsSegment"),
                    Ref("EngineTypeSegment"),
                    optional=True,
                ),
                # Key type
                Sequence(
                    OneOf(
                        Sequence("AGGREGATE", "KEY"),
                        Sequence("UNIQUE", "KEY"),
                        Sequence("PRIMARY", "KEY"),
                        Sequence("DUPLICATE", "KEY"),
                    ),
                    Bracketed(Delimited(Ref("ColumnReferenceSegment"))),
                    optional=True,
                ),
                Ref("CommentClauseSegment", optional=True),
                # Partitioning
                Ref("PartitionSegment", optional=True),
                # Distribution
                Ref("DistributionSegment", optional=True),
                # Order by
                Sequence(
                    "ORDER",
                    "BY",
                    Bracketed(Delimited(Ref("ColumnReferenceSegment"))),
                    optional=True,
                ),
                Sequence(
                    "PROPERTIES",
                    Bracketed(
                        Delimited(
                            Sequence(
                                Ref("QuotedLiteralSegment"),
                                Ref("EqualsSegment"),
                                Ref("QuotedLiteralSegment"),
                            )
                        )
                    ),
                    optional=True,
                ),
            ),
            # Create table like
            Sequence("LIKE", Ref("TableReferenceSegment")),
            # Create table as
            Sequence(
                Ref.keyword("AS", optional=True),
                OptionallyBracketed(Ref("SelectableGrammar")),
            ),
        ),
    )


class ColumnConstraintSegment(mysql.ColumnConstraintSegment):
    """A column option; each CREATE TABLE column can have 0 or more."""

    match_grammar: Matchable = OneOf(
        mysql.ColumnConstraintSegment.match_grammar,
        Sequence("AS", Ref("ExpressionSegment")),
    )


class PartitionSegment(BaseSegment):
    """A partition segment supporting StarRocks specific syntax.

    Supports three types of partitioning:
    1. Range partitioning (PARTITION BY RANGE)
    2. Expression partitioning using time functions (date_trunc/time_slice)
    3. Expression partitioning using column expressions
    """

    type = "partition_segment"
    match_grammar = Sequence(
        "PARTITION",
        "BY",
        OneOf(
            # Range partitioning
            Sequence(
                "RANGE",
                Bracketed(Delimited(Ref("ColumnReferenceSegment"))),
                OneOf(
                    # Fixed partitions
                    Bracketed(
                        Delimited(
                            Sequence(
                                "PARTITION",
                                Ref("ObjectReferenceSegment"),
                                "VALUES",
                                OneOf(
                                    # LESS THAN syntax
                                    Sequence(
                                        "LESS",
                                        "THAN",
                                        OneOf(
                                            "MAXVALUE",
                                            Bracketed(Delimited(Ref("LiteralGrammar"))),
                                        ),
                                    ),
                                    # Fixed range syntax
                                    Sequence(
                                        Bracketed(
                                            Bracketed(Delimited(Ref("LiteralGrammar"))),
                                            ",",
                                            Bracketed(Delimited(Ref("LiteralGrammar"))),
                                        )
                                    ),
                                ),
                            )
                        )
                    ),
                    # Dynamic partitions
                    Bracketed(
                        Sequence(
                            "START",
                            Bracketed(Ref("QuotedLiteralSegment")),
                            "END",
                            Bracketed(Ref("QuotedLiteralSegment")),
                            "EVERY",
                            Bracketed(
                                OneOf(
                                    Ref("QuotedLiteralSegment"),
                                    Sequence(
                                        "INTERVAL",
                                        Ref("NumericLiteralSegment"),
                                        OneOf("YEAR", "MONTH", "DAY", "HOUR"),
                                    ),
                                )
                            ),
                        )
                    ),
                ),
            ),
            # Expression partitioning - time function expressions
            Ref("FunctionSegment"),
            # Expression partitioning - column expressions
            Bracketed(Delimited(Ref("ColumnReferenceSegment"))),
        ),
    )


class DistributionSegment(BaseSegment):
    """A distribution segment supporting both hash and random distribution."""

    type = "distribution_segment"
    match_grammar = Sequence(
        "DISTRIBUTED",
        "BY",
        OneOf(
            Sequence(
                "HASH",
                Bracketed(Delimited(Ref("ColumnReferenceSegment"))),
                Sequence("BUCKETS", Ref("NumericLiteralSegment"), optional=True),
            ),
            Sequence(
                "RANDOM",
                Sequence("BUCKETS", Ref("NumericLiteralSegment"), optional=True),
            ),
        ),
    )


class QualifyClauseSegment(BaseSegment):
    """A `QUALIFY` clause like in `SELECT`.

    StarRocks QUALIFY constraints:
    - Only supports three window functions: ROW_NUMBER(), RANK(), and DENSE_RANK()
    - Execution order: FROM → WHERE → GROUP BY → HAVING →
        WINDOW → QUALIFY → DISTINCT → ORDER BY → LIMIT

    https://docs.starrocks.io/docs/sql-reference/sql-functions/Window_function/#qualify
    """

    type = "qualify_clause"
    match_grammar = Sequence(
        "QUALIFY",
        ImplicitIndent,
        # Must be a window function followed by comparison
        Sequence(
            Ref("FunctionSegment"),  # Window function with OVER clause
            Ref("ComparisonOperatorGrammar"),
            Ref("ExpressionSegment"),
        ),
        Dedent,
    )


class IndexDefinitionSegment(BaseSegment):
    """Bitmap index definition specific to StarRocks."""

    type = "index_definition"
    match_grammar = Sequence(
        "INDEX",
        Ref("IndexReferenceSegment"),
        Bracketed(Delimited(Ref("ColumnReferenceSegment"))),
        Sequence("USING", "BITMAP", optional=True),
        Sequence("COMMENT", Ref("QuotedLiteralSegment"), optional=True),
    )


class CreateRoutineLoadStatementSegment(BaseSegment):
    """A `CREATE ROUTINE LOAD` statement for StarRocks."""

    type = "create_routine_load_statement"
    match_grammar = Sequence(
        "CREATE",
        "ROUTINE",
        "LOAD",
        Ref("ObjectReferenceSegment"),  # job_name
        "ON",
        Ref("TableReferenceSegment"),  # table_name
        # Column definitions
        Sequence(
            "COLUMNS",
            Bracketed(
                Delimited(
                    OneOf(
                        Ref("QuotedIdentifierSegment"),
                        Ref("NakedIdentifierSegment"),
                        Sequence(
                            OneOf(
                                Ref("QuotedIdentifierSegment"),
                                Ref("NakedIdentifierSegment"),
                            ),
                            Ref("EqualsSegment"),
                            Ref("ExpressionSegment"),
                        ),
                    )
                )
            ),
            optional=True,
        ),
        # Properties section using dedicated properties segment
        Sequence(
            "PROPERTIES",
            Bracketed(Delimited(Ref("CreateRoutineLoadPropertiesSegment"))),
            optional=True,
        ),
        # Data Source section using dedicated data source properties segment
        "FROM",
        "KAFKA",
        Bracketed(Delimited(Ref("CreateRoutineLoadDataSourcePropertiesSegment"))),
    )


class CreateRoutineLoadPropertiesSegment(BaseSegment):
    """Properties segment for CREATE ROUTINE LOAD statement."""

    type = "routine_load_properties"
    match_grammar = Sequence(
        Ref("QuotedLiteralSegment"), Ref("EqualsSegment"), Ref("QuotedLiteralSegment")
    )


class CreateRoutineLoadDataSourcePropertiesSegment(BaseSegment):
    """Data source properties segment for CREATE ROUTINE LOAD statement."""

    type = "routine_load_data_source_properties"
    match_grammar = Sequence(
        Ref("QuotedLiteralSegment"), Ref("EqualsSegment"), Ref("QuotedLiteralSegment")
    )


class LoadLabelPropertySegment(BaseSegment):
    """A key/value property entry for LOAD LABEL and BROKER properties."""

    type = "load_label_property"
    match_grammar = Sequence(
        Ref("QuotedLiteralSegment"), Ref("EqualsSegment"), Ref("QuotedLiteralSegment")
    )


class LoadDataDescSegment(BaseSegment):
    """A data description entry for LOAD LABEL."""

    type = "load_data_desc"
    match_grammar = Sequence(
        "DATA",
        "INFILE",
        Bracketed(Delimited(Ref("QuotedLiteralSegment"))),
        "INTO",
        "TABLE",
        Ref("TableReferenceSegment"),
        Sequence(
            "FORMAT",
            "AS",
            Ref("QuotedLiteralSegment"),
            optional=True,
        ),
        Sequence(
            Bracketed(Delimited(Ref("ColumnReferenceSegment"))),
            optional=True,
        ),
    )


class LoadLabelStatementSegment(BaseSegment):
    """A `LOAD LABEL` statement for StarRocks."""

    type = "load_label_statement"
    match_grammar = Sequence(
        "LOAD",
        "LABEL",
        OneOf(
            Sequence(
                Ref("DatabaseReferenceSegment"),
                Ref("DotSegment"),
                Ref("ObjectReferenceSegment"),
            ),
            Ref("ObjectReferenceSegment"),
        ),
        Bracketed(Delimited(Ref("LoadDataDescSegment"))),
        "WITH",
        "BROKER",
        Bracketed(Delimited(Ref("LoadLabelPropertySegment"))),
        Sequence(
            "PROPERTIES",
            Bracketed(Delimited(Ref("LoadLabelPropertySegment"))),
            optional=True,
        ),
    )


"""Grammar for STOP ROUTINE LOAD statement in StarRocks."""


class StopRoutineLoadStatementSegment(BaseSegment):
    """A `STOP ROUTINE LOAD` statement.

    Stops a running routine load job.

    STOP ROUTINE LOAD FOR [db_name.]<job_name>
    """

    type = "stop_routine_load_statement"
    match_grammar = Sequence(
        "STOP",
        "ROUTINE",
        "LOAD",
        "FOR",
        OneOf(
            # db_name.job_name format
            Sequence(
                Ref("DatabaseReferenceSegment"),
                Ref("DotSegment"),
                Ref("ObjectReferenceSegment"),
            ),
            # job_name only format
            Ref("ObjectReferenceSegment"),
        ),
    )


class PauseRoutineLoadStatementSegment(BaseSegment):
    """A `PAUSE ROUTINE LOAD` statement.

    Pauses a running routine load job.
    """

    type = "pause_routine_load_statement"
    match_grammar = Sequence(
        "PAUSE",
        "ROUTINE",
        "LOAD",
        "FOR",
        Ref("ObjectReferenceSegment"),
    )


class ResumeRoutineLoadStatementSegment(BaseSegment):
    """A `RESUME ROUTINE LOAD` statement.

    Resumes a paused routine load job.
    """

    type = "resume_routine_load_statement"
    match_grammar = Sequence(
        "RESUME",
        "ROUTINE",
        "LOAD",
        "FOR",
        Ref("ObjectReferenceSegment"),
    )


class InsertOverwriteStatementSegment(BaseSegment):
    """A `INSERT OVERWRITE` statement with optional PARTITION and LABEL clauses."""

    type = "insert_overwrite_statement"
    match_grammar = Sequence(
        "INSERT",
        "OVERWRITE",
        Ref("TableReferenceSegment"),
        Sequence(
            "PARTITION",
            Bracketed(
                Delimited(
                    OneOf(
                        Ref("ExpressionSegment"),
                        Ref("NakedIdentifierSegment"),
                        Ref("ObjectReferenceSegment"),
                    )
                )
            ),
            optional=True,
        ),
        Sequence(
            "WITH",
            "LABEL",
            Ref("SingleIdentifierGrammar"),
            optional=True,
        ),
        Ref("SelectableGrammar"),
    )


class StatementSegment(mysql.StatementSegment):
    """Overriding StatementSegment to allow for additional segment parsing."""

    match_grammar = mysql.StatementSegment.match_grammar.copy(
        insert=[
            Ref("LoadLabelStatementSegment"),
            Ref("CreateRoutineLoadStatementSegment"),
            Ref("StopRoutineLoadStatementSegment"),
            Ref("PauseRoutineLoadStatementSegment"),
            Ref("ResumeRoutineLoadStatementSegment"),
        ]
    ).copy(
        insert=[
            Ref("InsertOverwriteStatementSegment"),
        ],
        at=1,  # ensure that it takes priority over MySQL's native InsertStatementSegment.
    )


class UnorderedSelectStatementSegment(mysql.UnorderedSelectStatementSegment):
    """A `SELECT` statement without any ORDER clauses or later.

    Enhanced for StarRocks to include QUALIFY clause support.
    QUALIFY is positioned after WINDOW clause per StarRocks execution order.
    """

    type = "select_statement"

    match_grammar = mysql.UnorderedSelectStatementSegment.match_grammar.copy(
        insert=[Ref("QualifyClauseSegment", optional=True)],
    )


class SelectStatementSegment(mysql.SelectStatementSegment):
    """A `SELECT` statement including QUALIFY support.

    StarRocks supports QUALIFY for filtering based on window functions.
    """

    type = "select_statement"

    match_grammar = UnorderedSelectStatementSegment.match_grammar.copy(
        insert=[
            Ref("OrderByClauseSegment", optional=True),
            Ref("LimitClauseSegment", optional=True),
            Ref("IntoClauseSegment", optional=True),
        ],
        terminators=[
            Ref("SetOperatorSegment"),
            Ref("UpsertClauseListSegment"),
            Ref("WithCheckOptionSegment"),
        ],
        replace_terminators=True,
    )



"""Implementation of Rule CP02."""

from typing import Optional

from sqlfluff.core.rules import LintResult, RuleContext
from sqlfluff.core.rules.crawlers import SegmentSeekerCrawler
from sqlfluff.rules.capitalisation.CP01 import Rule_CP01
from sqlfluff.utils.identifers import identifiers_policy_applicable


class Rule_CP02(Rule_CP01):
    """Inconsistent capitalisation of unquoted identifiers.

    This rule applies to all unquoted identifiers, whether references
    or aliases, and whether they refer to columns or other objects (such
    as tables or schemas).

    .. note::

       In **most** dialects, unquoted identifiers are treated as case-insensitive
       and so the fixes proposed by this rule do not change the interpretation
       of the query. **HOWEVER**, some databases (notably :ref:`bigquery_dialect_ref`,
       :ref:`trino_dialect_ref` and :ref:`clickhouse_dialect_ref`) do take the casing
       of *unquoted* identifiers into account when determining the casing of the column
       heading in the *result*.

       As this feature is only present in a few dialects, and not widely understood
       by users, we regard it as *an antipattern*. It is more widely understood that
       if the case of an identifier *matters*, then it should be quoted. If you, or
       your organisation, do wish to rely on this feature, we recommend that you
       disabled this rule (see :ref:`ruleselection`).

    **Anti-pattern**

    In this example, unquoted identifier ``a`` is in lower-case but
    ``B`` is in upper-case.

    .. code-block:: sql

        select
            a,
            B
        from foo

    In this more complicated example, there are a mix of capitalisations
    in both reference and aliases of columns and tables. That inconsistency
    is acceptable when those identifiers are quoted, but not when unquoted.

    .. code-block:: sql

        select
            col_1 + Col_2 as COL_3,
            "COL_4" as Col_5
        from Foo as BAR

    **Best practice**

    Ensure all unquoted identifiers are either in upper-case or in lower-case.

    .. code-block:: sql

        select
            a,
            b
        from foo;

        -- ...also good...

        select
            A,
            B
        from foo;

        --- ...or for comparison with our more complex example, this too:

        select
            col_1 + col_2 as col_3,
            "COL_4" as col_5
        from foo as bar

    """

    name = "capitalisation.identifiers"
    aliases = ("L014",)
    is_fix_compatible = True

    crawl_behaviour = SegmentSeekerCrawler(
        {"naked_identifier", "properties_naked_identifier"}
    )
    config_keywords = [
        "extended_capitalisation_policy",
        "unquoted_identifiers_policy",
        "ignore_words",
        "ignore_words_regex",
    ]
    _description_elem = "Unquoted identifiers"

    def _eval(self, context: RuleContext) -> Optional[list[LintResult]]:
        # Return None if identifier is case-sensitive property to enable Change
        # Data Feed
        # https://docs.delta.io/2.0.0/delta-change-data-feed.html#enable-change-data-feed
        if (
            context.dialect.name in ["databricks", "sparksql"]
            and context.parent_stack
            and context.parent_stack[-1].type == "property_name_identifier"
            and context.segment.raw == "enableChangeDataFeed"
        ):
            return None

        # Skip templated segments (e.g., placeholder parameters like :accountId)
        # to avoid incorrect qualification suggestions
        if context.segment.is_templated:
            return [LintResult(memory=context.memory)]

        if identifiers_policy_applicable(
            self.unquoted_identifiers_policy,  # type: ignore
            context.parent_stack,
        ):
            return super()._eval(context=context)
        else:
            return [LintResult(memory=context.memory)]



"""Common classes for dialects to use."""

import warnings
from collections.abc import Generator
from enum import Enum
from typing import NamedTuple, Optional, Protocol, Union, cast, runtime_checkable

from sqlfluff.core.parser import BaseSegment, RawSegment


class AliasInfo(NamedTuple):
    """Details about a table alias."""

    ref_str: str  # Name given to the alias
    segment: Optional[BaseSegment]  # Identifier segment containing the name
    aliased: bool
    from_expression_element: BaseSegment
    alias_expression: Optional[BaseSegment]
    object_reference: Optional[BaseSegment]


class ColumnAliasInfo(NamedTuple):
    """Details about a column alias."""

    alias_identifier_name: str
    aliased_segment: BaseSegment
    column_reference_segments: list[BaseSegment]


class ObjectReferencePart(NamedTuple):
    """Details about a part of an object reference."""

    part: str  # Name of the part
    # Segment(s) comprising the part. Usually just one segment, but could
    # be multiple in dialects (e.g. BigQuery) that support unusual
    # characters in names (e.g. "-")
    segments: list[RawSegment]


@runtime_checkable
class _SupportsIterRawReferences(Protocol):
    """An object reference segment exposing the legacy ``iter_raw_references``.

    Used to type (and guard) the back-compat ``dialect_name=None`` dispatch path
    without importing the concrete dialect segment classes -- doing so would be a
    circular import and a forbidden ``core -> dialects`` dependency.
    """

    def iter_raw_references(self) -> Generator[ObjectReferencePart, None, None]: ...


class ObjectReferenceLevel(Enum):
    """Labels for the "levels" of a reference.

    Note: Since SQLFluff does not have access to database catalog
    information, interpreting references will often be ambiguous.
    Typical example: The first part *may* refer to a schema, but that is
    almost always optional if referring to an object in some default or
    currently "active" schema. For this reason, use of this enum is optional
    and intended mainly to clarify the intent of the code -- no guarantees!
    Additionally, the terminology may vary by dialect, e.g. in BigQuery,
    "project" would be a more accurate term than "schema".
    """

    OBJECT = 1
    TABLE = 2
    SCHEMA = 3


def _level_to_int(level: Union[ObjectReferenceLevel, int]) -> int:
    # If it's an ObjectReferenceLevel, get the value. Otherwise, assume it's
    # an int.
    level = getattr(level, "value", level)
    assert isinstance(level, int)
    return level


# =============================================================================
# Functional helpers for object references.
#
# These mirror methods that historically lived on ObjectReferenceSegment (and
# its dialect-specific subclasses). They are defined here as free functions
# that dispatch on the dialect name + segment type so that the logic is
# portable (e.g. to the Rust parser, where segments are generic and carry only
# their type information plus a dialect name). The corresponding segment
# methods are retained as deprecated thin wrappers that delegate here.
# =============================================================================


def deprecated_segment_method(old: str, new: str) -> None:
    """Warn that a reference/alias segment method is deprecated.

    These methods have been replaced by the free functions in this module so
    that the logic can be dispatched on the dialect name + segment type (rather
    than the Python segment class), which is a prerequisite for porting the
    rules that use them to Rust. Call this from the deprecated wrapper methods.
    """
    warnings.warn(
        f"{old} is deprecated and will be removed in a future release. "
        f"Use sqlfluff.core.dialects.common.{new} instead.",
        DeprecationWarning,
        stacklevel=3,
    )


def _iter_raw_references_default(
    segment: BaseSegment, *, split_on_dots: bool
) -> Generator[ObjectReferencePart, None, None]:
    """Yield the reference parts of a generic object reference.

    When ``split_on_dots`` is True (e.g. BigQuery's splittable references), each
    identifier is further split on embedded dots.
    """
    for elem in segment.recursive_crawl("identifier"):
        raw_trimmed = cast(RawSegment, elem).raw_trimmed()
        if split_on_dots:
            for part in raw_trimmed.split("."):
                yield ObjectReferencePart(part, [cast(RawSegment, elem)])
        else:
            yield ObjectReferencePart(raw_trimmed, [cast(RawSegment, elem)])


def _iter_raw_references_wildcard(
    segment: BaseSegment,
) -> Generator[ObjectReferencePart, None, None]:
    """Yield the reference parts of a wildcard identifier (includes ``star``)."""
    for elem in segment.recursive_crawl("identifier", "star"):
        yield ObjectReferencePart(
            cast(RawSegment, elem).raw_trimmed(), [cast(RawSegment, elem)]
        )


def _iter_raw_references_bigquery_table(
    segment: BaseSegment,
) -> Generator[ObjectReferencePart, None, None]:
    """Yield reference parts for a BigQuery table reference.

    Overrides the default because hyphens (DashSegment) cause one logical part
    of the name to be split across multiple elements, e.g. "table-a" is parsed
    as three segments.
    """
    # For each descendant element, group them, using "dot" elements as a
    # delimiter.
    parts: list[str] = []
    elems_for_parts: list[RawSegment] = []

    def flush() -> ObjectReferencePart:
        nonlocal parts, elems_for_parts
        result = ObjectReferencePart("".join(parts), elems_for_parts)
        parts = []
        elems_for_parts = []
        return result

    for elem in segment.recursive_crawl("identifier", "literal", "dash", "dot", "star"):
        if not elem.is_type("dot"):
            if elem.is_type("identifier"):
                # Found an identifier (potentially with embedded dots).
                elem_subparts = cast(RawSegment, elem).raw_trimmed().split(".")
                for idx, part in enumerate(elem_subparts):
                    # Save each part of the segment.
                    parts.append(part)
                    elems_for_parts.append(cast(RawSegment, elem))

                    if idx != len(elem_subparts) - 1:
                        # For each part except the last, flush.
                        yield flush()
            else:
                # For non-identifier segments, save the whole segment.
                parts.append(cast(RawSegment, elem).raw_trimmed())
                elems_for_parts.append(cast(RawSegment, elem))
        else:
            yield flush()

    # Flush any leftovers.
    if parts:
        yield flush()


def iter_raw_references(
    segment: BaseSegment, dialect_name: Optional[str]
) -> Generator[ObjectReferencePart, None, None]:
    """Generate the reference parts of an object reference.

    Dispatches on the dialect name and segment type to reproduce the behavior
    previously provided by dialect-specific segment subclasses.

    ``dialect_name`` of None is a back-compat path for the deprecated segment
    method wrappers (which have no dialect to hand): it dispatches via the
    segment's own (deprecated) ``iter_raw_references`` method so that Python
    class dispatch -- and therefore the exact legacy behavior -- still applies.
    Always pass the real dialect name from non-deprecated call sites.
    """
    if dialect_name is None:
        assert isinstance(segment, _SupportsIterRawReferences), (
            "iter_raw_references(segment, None) requires an object reference "
            f"segment (with a legacy iter_raw_references method); got {segment.type!r}"
        )
        yield from segment.iter_raw_references()
        return
    if segment.is_type("wildcard_identifier"):
        yield from _iter_raw_references_wildcard(segment)
        return
    if dialect_name == "bigquery":
        if segment.is_type("table_reference"):
            yield from _iter_raw_references_bigquery_table(segment)
            return
        if segment.is_type("column_reference"):
            yield from _iter_raw_references_default(segment, split_on_dots=True)
            return
    yield from _iter_raw_references_default(segment, split_on_dots=False)


def _raw_refs(
    segment: BaseSegment, dialect_name: Optional[str]
) -> list[ObjectReferencePart]:
    """Collect the raw reference parts of a segment."""
    return list(iter_raw_references(segment, dialect_name))


def is_qualified(segment: BaseSegment, dialect_name: Optional[str]) -> bool:
    """Return whether there is more than one element to the reference."""
    return len(_raw_refs(segment, dialect_name)) > 1


def qualification(segment: BaseSegment, dialect_name: Optional[str]) -> str:
    """Return the qualification type of this reference."""
    return "qualified" if is_qualified(segment, dialect_name) else "unqualified"


def _extract_possible_references_default(
    segment: BaseSegment,
    level: Union[ObjectReferenceLevel, int],
    dialect_name: Optional[str],
) -> list[ObjectReferencePart]:
    """Extract possible references of a given level."""
    level_int = _level_to_int(level)
    refs = _raw_refs(segment, dialect_name)
    if len(refs) >= level_int:
        return [refs[-level_int]]
    return []


def _extract_possible_references_bigquery_column(
    segment: BaseSegment,
    level: Union[ObjectReferenceLevel, int],
    dialect_name: Optional[str],
) -> list[ObjectReferencePart]:
    """Extract possible references of a given level for a BigQuery column.

    BigQuery's support for things like the following:
    - Functions that take a table as a parameter (e.g. TO_JSON_STRING)
      https://cloud.google.com/bigquery/docs/reference/standard-sql/
      json_functions#to_json_string
    - STRUCT

    means that, without schema information (which SQLFluff does not have),
    references to data are often ambiguous.
    """
    level_int = _level_to_int(level)
    refs = _raw_refs(segment, dialect_name)
    if level_int == ObjectReferenceLevel.SCHEMA.value and len(refs) >= 3:
        return [refs[0]]  # pragma: no cover
    if level_int == ObjectReferenceLevel.TABLE.value:
        # One part: Could be a table, e.g. TO_JSON_STRING(t)
        # Two parts: Could be dataset.table or table.column.
        # Three parts: Could be table.column.struct or dataset.table.column.
        # Four parts: dataset.table.column.struct
        # Five parts: project.dataset.table.column.struct
        # So... return the first 3 parts.
        return refs[:3]
    if (
        level_int == ObjectReferenceLevel.OBJECT.value and len(refs) >= 3
    ):  # pragma: no cover
        # Ambiguous case: The object (i.e. column) could be the first or
        # second part, so return both.
        return [refs[1], refs[2]]
    return _extract_possible_references_default(
        segment, level, dialect_name
    )  # pragma: no cover


def extract_possible_references(
    segment: BaseSegment,
    level: Union[ObjectReferenceLevel, int],
    dialect_name: Optional[str],
) -> list[ObjectReferencePart]:
    """Extract possible references of a given level.

    "level" may be (but is not required to be) a value from the
    ObjectReferenceLevel enum defined above.

    NOTE: The base implementation here returns at most one part, but
    dialects such as BigQuery that support nesting (e.g. STRUCT) may return
    multiple reference parts.
    """
    if dialect_name == "bigquery" and segment.is_type("column_reference"):
        return _extract_possible_references_bigquery_column(
            segment, level, dialect_name
        )
    return _extract_possible_references_default(segment, level, dialect_name)


def _extract_possible_multipart_references_default(
    segment: BaseSegment,
    levels: list[Union[ObjectReferenceLevel, int]],
    dialect_name: Optional[str],
) -> list[tuple[ObjectReferencePart, ...]]:
    """Extract possible multipart references, e.g. schema.table."""
    levels_tmp = [_level_to_int(level) for level in levels]
    min_level = min(levels_tmp)
    max_level = max(levels_tmp)
    refs = _raw_refs(segment, dialect_name)
    if len(refs) >= max_level:
        return [tuple(refs[-max_level : 1 - min_level])]
    return []


def _extract_possible_multipart_references_bigquery_column(
    segment: BaseSegment,
    levels: list[Union[ObjectReferenceLevel, int]],
    dialect_name: Optional[str],
) -> list[tuple[ObjectReferencePart, ...]]:
    """Extract possible multipart references for a BigQuery column."""
    levels_tmp = [_level_to_int(level) for level in levels]
    min_level = min(levels_tmp)
    max_level = max(levels_tmp)
    refs = _raw_refs(segment, dialect_name)
    if max_level == ObjectReferenceLevel.SCHEMA.value and len(refs) >= 3:
        return [tuple(refs[0 : max_level - min_level + 1])]
    # Note we aren't handling other possible cases. We'll add these as
    # needed.
    return _extract_possible_multipart_references_default(segment, levels, dialect_name)


def extract_possible_multipart_references(
    segment: BaseSegment,
    levels: list[Union[ObjectReferenceLevel, int]],
    dialect_name: Optional[str],
) -> list[tuple[ObjectReferencePart, ...]]:
    """Extract possible multipart references, e.g. schema.table."""
    if dialect_name == "bigquery" and segment.is_type("column_reference"):
        return _extract_possible_multipart_references_bigquery_column(
            segment, levels, dialect_name
        )
    return _extract_possible_multipart_references_default(segment, levels, dialect_name)


# =============================================================================
# Functional helpers for "eventual" aliases of FROM / JOIN segments.
# =============================================================================


def get_from_expression_element_alias(
    segment: BaseSegment, dialect_name: Optional[str]
) -> Generator[AliasInfo, None, None]:
    """Return the eventual table name referred to by a table expression.

    ``segment`` should be a ``from_expression_element``.
    """
    # Get any table expressions
    tbl_expression = segment.get_child("table_expression")
    if not tbl_expression:  # pragma: no cover
        _bracketed = segment.get_child("bracketed")
        if _bracketed:
            tbl_expression = _bracketed.get_child("table_expression")
    # For TSQL nested, bracketed tables get the first table as reference
    if tbl_expression and not tbl_expression.get_child("object_reference"):
        _bracketed = tbl_expression.get_child("bracketed")
        if _bracketed:
            tbl_expression = _bracketed.get_child("table_expression")

    # Work out the references
    ref: Optional[BaseSegment] = None
    if tbl_expression:
        ref = tbl_expression.get_child("object_reference")

    # Handle any aliases
    has_alias = False
    alias_expressions = segment.get_children("alias_expression", "bracketed")
    for alias_expression in alias_expressions:
        if alias_expression.is_type("bracketed"):  # pragma: no cover
            _alias_expression = alias_expression.get_child("alias_expression")
            if _alias_expression is None:
                continue
            alias_expression = _alias_expression
        # If it has an alias, return that
        has_alias = True
        alias_segment = alias_expression.get_child("identifier")
        if alias_segment:
            yield AliasInfo(
                alias_segment.raw_normalized(casefold=False),
                alias_segment,
                True,
                segment,
                alias_expression,
                ref,
            )
    if has_alias:
        return

    # If not return the object name (or None if there isn't one)
    if ref:
        references = list(iter_raw_references(ref, dialect_name))
        # Return the last element of the reference.
        if references:
            penultimate_ref = references[-1]
            yield AliasInfo(
                penultimate_ref.part,
                penultimate_ref.segments[0],
                False,
                segment,
                None,
                ref,
            )
            return
    # No references or alias
    yield AliasInfo(
        "",
        None,
        False,
        segment,
        None,
        ref,
    )


def get_join_clause_aliases(
    segment: BaseSegment, dialect_name: Optional[str]
) -> list[tuple[BaseSegment, AliasInfo]]:
    """Return the eventual table name referred to by a join clause."""
    buff = []

    from_expression = segment.get_child("from_expression_element")
    # As per grammar, there will always be a FromExpressionElementSegment
    assert from_expression
    from_aliases = get_from_expression_element_alias(from_expression, dialect_name)
    # Only append if non-null. A None reference, may
    # indicate a generator expression or similar.
    for alias in from_aliases:
        buff.append((from_expression, alias))

    # In some dialects, like TSQL, join clauses can have nested join clauses
    # recurse into them - but not if part of a sub-select statement (see #3144)
    for join_clause in segment.recursive_crawl(
        "join_clause", no_recursive_seg_type="select_statement"
    ):
        if join_clause is segment:
            # If the starting segment itself matches the list of types we're
            # searching for, recursive_crawl() will return it. Skip that.
            continue
        aliases = get_join_clause_aliases(join_clause, dialect_name)
        # Only append if non-null. A None reference, may
        # indicate a generator expression or similar.
        if aliases:
            buff = buff + aliases
    return buff


def get_from_clause_aliases(
    segment: BaseSegment, dialect_name: Optional[str]
) -> list[tuple[BaseSegment, AliasInfo]]:
    """List the eventual aliases of a from clause.

    Comes as a list of tuples (table expr, tuple (string, segment, bool)).
    """
    buff: list[tuple[BaseSegment, AliasInfo]] = []
    direct_table_children = []
    join_clauses = []

    for from_expression in segment.get_children("from_expression"):
        direct_table_children += from_expression.get_children("from_expression_element")
        join_clauses += from_expression.get_children("join_clause")

    # Iterate through the potential sources of aliases
    for clause in direct_table_children:
        direct_table_aliases = get_from_expression_element_alias(clause, dialect_name)
        # Only append if non-null. A None reference, may
        # indicate a generator expression or similar.
        table_expr = (
            clause
            if clause in direct_table_children
            else clause.get_child("from_expression_element")
        )
        for alias in direct_table_aliases:
            assert table_expr
            buff.append((table_expr, alias))
    for clause in join_clauses:
        aliases = get_join_clause_aliases(clause, dialect_name)
        # Only append if non-null. A None reference, may
        # indicate a generator expression or similar.
        if aliases:
            buff = buff + aliases
    return buff



"""A List of FlinkSQL keywords.

Based on Apache Flink SQL reference:
https://nightlies.apache.org/flink/flink-docs-release-1.18/docs/dev/table/sql/queries/
"""

# FlinkSQL Reserved Keywords
RESERVED_KEYWORDS = [
    "ALL",
    "AND",
    "ANY",
    "AS",
    "AUTHORIZATION",
    "BETWEEN",
    "BIGINT",
    "BINARY",
    "BOOLEAN",
    "BOTH",
    "BY",
    "CASE",
    "CAST",
    "CHAR",
    "CHARACTER",
    "CHECK",
    "COLLATE",
    "COLUMN",
    "CONSTRAINT",
    "CREATE",
    "CROSS",
    "CURRENT_DATE",
    "CURRENT_TIME",
    "CURRENT_TIMESTAMP",
    "CURRENT_USER",
    "CURSOR",
    "DATE",
    "DAY",
    "DECIMAL",
    "DECLARE",
    "DELETE",
    "DESC",
    "DISTINCT",
    "DOUBLE",
    "DROP",
    "ELSE",
    "END",
    "ESCAPE",
    "EXCEPT",
    "EXISTS",
    "EXTRACT",
    "FALSE",
    "FETCH",
    "FILTER",
    "FLOAT",
    "FOR",
    "FOREIGN",
    "FROM",
    "FULL",
    "FUNCTION",
    "GRANT",
    "GROUP",
    "HAVING",
    "HOUR",
    "IF",
    "IN",
    "INNER",
    "INSERT",
    "INT",
    "INTEGER",
    "INTERSECT",
    "INTERVAL",
    "INTO",
    "IS",
    "JOIN",
    "LEADING",
    "LEFT",
    "LIKE",
    "LIMIT",
    "LOCAL",
    "MINUTE",
    "MONTH",
    "NATURAL",
    "NOT",
    "NULL",
    "NUMERIC",
    "OF",
    "ON",
    "ONLY",
    "OR",
    "ORDER",
    "OUTER",
    "OVERLAPS",
    "OVERLAY",
    "PARTITION",
    "POSITION",
    "PRIMARY",
    "REAL",
    "REFERENCES",
    "RIGHT",
    "ROW",
    "SECOND",
    "SELECT",
    "SESSION_USER",
    "SET",
    "SMALLINT",
    "SOME",
    "SUBSTRING",
    "TABLE",
    "THEN",
    "TIME",
    "TIMESTAMP",
    "TINYINT",
    "TO",
    "TRAILING",
    "TRUE",
    "UNION",
    "UNIQUE",
    "UNKNOWN",
    "UPDATE",
    "USER",
    "USING",
    "VALUES",
    "VARCHAR",
    "WHEN",
    "WHERE",
    "WITH",
    "YEAR",
]

# FlinkSQL Unreserved Keywords
UNRESERVED_KEYWORDS = [
    "ADD",
    "AFTER",
    "ANALYZE",
    "APPEND",
    "ARRAY",
    "ASC",
    "AVRO_CONFLUENT",
    "AVRO",
    "BEGIN",
    "BIGQUERY",
    "BYTES",
    "CALL",
    "CANAL",
    "CANCEL",
    "CASSANDRA",
    "CATALOG",
    "CATALOGS",
    "CHANGELOG",
    "CHECKPOINT",
    "COMMENT",
    "COMPACTION",
    "COMPUTED",
    "CONNECTOR",
    "CSV",
    "CURRENT",
    "DATABASE",
    "DATABASES",
    "DEBEZIUM",
    "DESCRIBE",
    "DISTRIBUTED",
    "DISTRIBUTION",
    "ELASTIC",
    "ELASTICSEARCH",
    "ENFORCED",
    "EXCLUDING",
    "EXIT",
    "EXPLAIN",
    "EXTENDED",
    "FILESYSTEM",
    "FIRST",
    "FORMAT",
    "FUNCTIONS",
    "GENERATED",
    "HASH",
    "HBASE",
    "HELP",
    "HISTORY",
    "INCLUDING",
    "JARS",
    "JDBC",
    "JOBS",
    "JSON",
    "KAFKA",
    "KINESIS",
    "LAST",
    "LOAD",
    "MAP",
    "MAXWELL",
    "MERGE",
    "METADATA",
    "MODULES",
    "MULTISET",
    "OGG",
    "OPTION",
    "OPTIONS",
    "ORC",
    "OVERWRITING",
    "PARQUET",
    "PARTITIONED",
    "PARTITIONS",
    "PHYSICAL",
    "PLAN",
    "PROCEDURE",
    "PROCEDURES",
    "PROTOBUF_CONFLUENT",
    "PROTOBUF",
    "PULSAR",
    "QUIT",
    "RANGE",
    "RAW",
    "REDIS",
    "REPLACE",
    "RESET",
    "RESTART",
    "RESUME",
    "RETRACT",
    "SAVEPOINT",
    "SET",
    "SHOW",
    "SINK",
    "SOURCE",
    "START",
    "STOP",
    "SUSPEND",
    "SYSTEM",
    "TABLES",
    "TEMPORARY",
    "TIMESTAMP_LTZ",
    "TRUNCATE",
    "UNLOAD",
    "UPSERT",
    "USE",
    "VALUE",
    "VIRTUAL",
    "WATERMARK",
    "WATERMARKS",
    "YEAR",
    "ZONE",
]



"""Common classes and functions for defining templating builtins."""

from typing import Any, Callable

from sqlfluff.core.errors import SQLTemplaterError


class FunctionWrapper:
    """Class to wrap a callable, for better error handling.

    When called, it just delegates to the provided callable, but if
    it is rendered as a string directly, it generates a templating
    error.
    """

    def __init__(self, name: str, callable: Callable[..., Any]):
        self._name = name
        self._callable = callable

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        """When the wrapper is called, call the internal function."""
        return self._callable(*args, **kwargs)

    def __str__(self) -> str:
        """If we try and render the wrapper directly, throw an error."""
        raise SQLTemplaterError(
            f"Unable to render builtin callable {self._name!r} as a "
            "variable because it is defined as a function. To remove "
            "this function from the context, set `apply_dbt_builtins` "
            "to False."
        )
